"""

    SPI Bus
    This class is base on:
    https://github.com/schang412/cocotbext-spi/blob/main/cocotbext/spi/spi.py

"""
import cocotb
import random
import logging

from cocotb.clock import Clock, BaseClock
from cocotb.triggers import RisingEdge, FallingEdge, Timer, Event, Edge
from cocotb.result import *
from cocotb.utils import *
from cocotb.queue import Queue, QueueFull

# SPI
class SpiDevice(object):

    def __init__(self, dut, sclk, mosi, miso, cs):
        self.dut = dut
        self.SCLK = sclk
        self.MOSI = mosi
        self.MISO = miso
        self.CS = cs
        self.CPOL = 0
        self.CPHA = 0
        self.SpiFreqz = 0.0

    def ConfigBus(self, SpiFreqz = 25000000.0, CPOL = False, CPHA = False, word_width=32):
        self.SpiFreqz = SpiFreqz
        self.CPOL = CPOL
        self.CPHA = CPHA
        self.word_width = word_width

class SpiClock(object):

    def __init__(self, signal, period, units="step", start_high=True):
        BaseClock.__init__(self, signal)
        self.period = cocotb.utils.get_sim_steps(period, units)
        self.half_period = cocotb.utils.get_sim_steps(period / 2.0, units)
        self.frequency = 1.0 / cocotb.utils.get_time_from_sim_steps(self.period, units='us')

        self.signal = signal

        self.start_high = start_high

        self._idle = Event()
        self._sync = Event()
        self._start = Event()

        self._idle.set()

        self._run_coroutine_obj = None
        self._restart()

    def _restart(self):
        if self._run_coroutine_obj is not None:
            self._run_cr.kill()
        self._run_cr = cocotb.start_soon(self._run())

    async def stop(self):
        self.stop_no_wait()
        await self._idle.wait()

    def stop_no_wait(self):
        self._start.clear()
        self._sync.set()

    async def start(self):
        self.start_no_wait()

    def start_no_wait(self):
        self._start.set()
        self._sync.set()

    async def _run(self):
        t = Timer(self.half_period)
        if self.start_high:
            while True:
                while not self._start.is_set():
                    self._idle.set()
                    self._sync.clear()
                    await self._sync.wait()

                self._idle.clear()
                self.signal.value = 1
                await t
                if self._start.is_set():
                    self.signal.value = 0
                    await t
        else:
            while True:
                while not self._start.is_set():
                    self._idle.set()
                    self._sync.clear()
                    await self._sync.wait()

                self._idle.clear()
                self.signal.value = 0
                await t
                if self._start.is_set():
                    self.signal.value = 1
                    await t

class SpiMaster(SpiDevice):

    def __init__(self, dut, sclk, mosi, miso, cs, SpiFreqz = 25000000.0, CPOL = False, CPHA = False, word_width=32):
        # Init Bus
        SpiDevice.__init__(self, dut=dut, sclk=sclk, mosi=mosi, miso=miso, cs=cs)
        # Config Bus
        SpiDevice.ConfigBus(self, SpiFreqz = SpiFreqz, CPOL = CPOL, CPHA = CPHA, word_width=word_width)

        self.SCLK.setimmediatevalue(self.CPOL)
        self.MOSI.setimmediatevalue(1)
        self.CS.setimmediatevalue(1)

        self.tx_fifo = Queue()
        self.rx_fifo = Queue()

        self.SpiClock = SpiClock(signal=self.SCLK, period=(1 / self.SpiFreqz), units="sec", start_high=self.CPHA)

        self.sync = Event()
        self._idle = Event()
        self._idle.set()

        self._run_coroutine_obj = None
        self.restart()

    def restart(self):
        if self._run_coroutine_obj is not None:
            self._run_coroutine_obj.kill()
        self._run_coroutine_obj = cocotb.start_soon(self.RunMaster())

    def count_tx(self):
        return self.tx_fifo.qsize()

    def check_tx_fifo_empty(self):
        return self.tx_fifo.empty()

    def count_rx(self):
        return self.rx_fifo.qsize()

    def check_rx_fifo_empty(self):
        return self.rx_fifo.empty()

    def idle(self):
        return self.check_tx_fifo_empty() and self.check_rx_fifo_empty()

    async def write(self, data):
        self.write_nowait(data)
        await self._idle.wait()

    def write_nowait(self, data):
        for b in data:
            self.tx_fifo.put_nowait(int(b))
        self.sync.set()
        self._idle.clear()

    async def read(self, count=-1):
        while self.check_rx_fifo_empty():
            self.sync.clear()
            await self.sync.wait()
        return self.read_nowait(count)

    def read_nowait(self, count=-1):
        if count < 0:
            count = self.count_rx()
        if self.word_width == 8:
            data = bytearray()
        else:
            data = []
        for k in range(count):
            data.append(self.rx_fifo.get_nowait())
        return data

    async def RunMaster(self):
        while True:
            TX_STARTED = False
            while self.check_tx_fifo_empty():
                self.SCLK.value = int(self.CPOL)
                self._idle.set()
                self.sync.clear()
                await self.sync.wait()

            while not (self.check_tx_fifo_empty()):
                tx_word = self.tx_fifo.get_nowait()
                rx_word = 0

                # the timing diagrams are CPHA/CPOL convention come from
                # https://en.wikipedia.org/wiki/Serial_Peripheral_Interface
                # this is also compliant with Linux Kernel definiton of SPI

                if TX_STARTED == False:
                    # if CPHA=0, the first bit is typically clocked out on edge of chip select
                    if not self.CPHA:
                        self.MOSI.value = bool(tx_word & (1 << self.word_width - 1))

                    # set the chip select
                    self.CS.value = int(False)
                    await Timer(self.SpiClock.period, units='step')

                    await self.SpiClock.start()
                    TX_STARTED = True

                if self.CPHA:
                    # if CPHA=1, the first edge is propagate, the second edge is sample
                    for k in range(self.word_width):
                        # the out changes on the leading edge of clock
                        await Edge(self.SCLK)
                        self.MOSI.value = bool(tx_word & (1 << (self.word_width - 1 - k)))

                        # while the in captures on the trailing edge of the clock
                        await Edge(self.SCLK)
                        rx_word |= bool(self.MISO.value.integer) << (self.word_width - 1 - k)
                else:
                    # if CPHA=0, the first edge is sample, the second edge is propagate
                    # we already clocked out one bit on edge of chip select, so we will clock out less bits
                    for k in range(self.word_width - 1):
                        await Edge(self.SCLK)
                        rx_word |= bool(self.MISO.value.integer) << (self.word_width - 1 - k)
                        await Edge(self.SCLK)
                        self.MOSI.value = bool(tx_word & (1 << (self.word_width - 2 - k)))
                    # but we haven't sampled enough times, so we will wait for another edge to sample
                    await Edge(self.SCLK)
                    rx_word |= bool(self.MISO.value.integer)

                self.rx_fifo.put_nowait(rx_word)

            # set sclk back to idle state
            await self.SpiClock.stop()
            self.SCLK.value = self.CPOL

            # wait another sclk period before restoring the chip select and mosi to idle (not necessarily part of spec)
            await Timer(self.SpiClock.period, units='step')
            self.CS.value = 1
            self.MOSI.value = 1

            self.sync.set()
