"""

    Test SPI SLAVE

"""
import cocotb
import numpy as np
import time
import random
import math

from cocotb.clock import Clock
from cocotb.triggers import RisingEdge, FallingEdge, Timer
from cocotb.regression import TestFactory
from cocotb.result import *
from spi import SpiClock, SpiMaster

# ------------------------------------------------------------------------------
# Some functions to support the test
# Define SPI mode
def define_spi_mode(dut, g_CPOL=0, g_CPHA=0):
    CPOL = False
    CPHA = False

    if g_CPOL == 1:
        CPOL = True
    if g_CPHA == 1:
        CPHA = True

    dut._log.info("     +-------------+")
    dut._log.info("     | CPOL | CPHA |")
    dut._log.info("     +-------------+")
    dut._log.info("MODE |  {}   |  {}   |".format(g_CPOL,g_CPHA))
    dut._log.info("     +-------------+")
    return CPOL, CPHA

# ------------------------------------------------------------------------------
# fill data
def fill_array(data_len, data_width, randomize):
    data = []
    mask = 2**data_width - 1
    for i in range(data_len):
        if randomize:
            data.append(random.randint(0, (2**data_width)-1) & mask)
        else:
            if i % 4 == 0:
                data.append(0xAAAAAAAA & mask)
            elif i % 4 == 1:
                data.append(0x33333333 & mask)
            elif i % 4 == 2:
                data.append(0xDEADCAFE & mask)
            elif i % 4 == 3:
                data.append(0xBADC0FFE & mask)
    data.append(0)
    return data

# ------------------------------------------------------------------------------
# Setup clock signal
def setup_clock(signal_name, period, units="ns"):
    cocotb.start_soon(Clock(signal_name, period, units=units).start())

# ------------------------------------------------------------------------------
# Main Test
@cocotb.coroutine
async def test_loopbackmode(dut, n_rst_cycles, data_len, sys_per, spi_freq, randomize):

    # Get data to send
    data_width = dut.g_data_width.value.integer
    g_cpol = dut.g_cpol.value.integer
    g_cpha = dut.g_cpha.value.integer

    CPOL, CPHA = define_spi_mode(dut=dut, g_CPOL=g_cpol, g_CPHA=g_cpha)

    data_wr = fill_array(data_len, data_width, randomize)

    spi_master = SpiMaster(
                            dut=dut,
                            sclk=dut.spi_clk_i,
                            mosi=dut.spi_mosi_i,
                            miso=dut.spi_miso_o,
                            cs=dut.spi_cs_n_i,
                            SpiFreqz = spi_freq,
                            CPOL = CPOL,
                            CPHA = CPHA,
                            word_width=data_width
                        )

    setup_clock(signal_name=dut.clk_i,period=sys_per)
    cocotb.start_soon(reset(dut, n_rst_cycles))
    await delay(dut.clk_i, 1)
    while dut.rst_n_i.value != 1:
        await RisingEdge(dut.clk_i)
    cocotb.start_soon(loopback(dut))

    data_written = cocotb.start_soon(write_validation(dut, data_wr))
    await spi_master.write(data_wr)
    data_read = await spi_master.read()

    if len(data_wr) != len(data_read):
        dut._log.error("Expected length {} --- Written length {}".format(len(data_wr),len(data_read)))
        dut._log.error("Write data failed, length is wrong in the SLAVE")
        raise SimFailure()
    else:
        dut._log.info("Data length READ is correct")
        data_read.pop(0)
        data_read.append(0)
        if (np.array_equal(data_wr, data_read)):
            dut._log.info("Data integrity READ is correct")
        else:
            dut._log.warning("Data WR: {}".format(data_wr))
            dut._log.warning("Data RD: {}".format(data_read))
            dut._log.error("Read data failed")
            raise SimFailure()

    await delay(dut.clk_i, 100)

    dut._log.info("Testbench finished successfully")

# ------------------------------------------------------------------------------
# Factory
factory = TestFactory(test_loopbackmode)
factory.add_option("n_rst_cycles", [11])
factory.add_option("data_len", [4,16,32,256])
factory.add_option("sys_per", [16])
factory.add_option("spi_freq", [1e6, 2e6, 5e6])
factory.add_option("randomize", [False, True])
factory.generate_tests()

# ------------------------------------------------------------------------------
# Reset
@cocotb.coroutine
def reset(dut, n_rst_cycles):
    dut._log.info("Asserting Sytem Reset")
    dut.rst_n_i.setimmediatevalue(0)
    dut.data_i.setimmediatevalue(0)
    dut.data_valid_i.setimmediatevalue(0)
    for i in range(n_rst_cycles):
        yield RisingEdge(dut.clk_i)
    dut.rst_n_i.value = 1
    yield RisingEdge(dut.clk_i)
    dut._log.info("Deasserting Sytem Reset")

# ------------------------------------------------------------------------------
# delay
@cocotb.coroutine
def delay(signal_name, n_cycles):
    for i in range(n_cycles):
        yield RisingEdge(signal_name)

# ------------------------------------------------------------------------------
# sample write side
@cocotb.coroutine
def write_validation(dut, data_wr):

    data_written = []
    while dut.spi_cs_n_i.value == 1:
        yield RisingEdge(dut.clk_i)
    while dut.spi_cs_n_i.value == 0:
        yield RisingEdge(dut.data_valid_o)
        data_written.append(dut.data_o.value.integer)

    if len(data_wr) != len(data_written):
        dut._log.error("Expected length {} --- Written length {}".format(len(data_wr),len(data_written)))
        dut._log.error("Write data failed, length is wrong in the SLAVE")
        raise SimFailure()
    else:
        dut._log.info("Data length is consistent in the SLAVE")
        if (np.array_equal(data_wr, data_written)):
            dut._log.info("Data integrity is correct in the SLAVE")
        else:
            dut._log.warning("Data WR     : {}".format(data_wr))
            dut._log.warning("Data Written: {}".format(data_written))
            dut._log.error("Write data failed, data is not the same in the SLAVE")
            raise SimFailure()

# ------------------------------------------------------------------------------
# Loopback
@cocotb.coroutine
def loopback(dut):
    while True:
        if dut.ready_o.value == 1 and dut.data_valid_o.value == 1:
            dut.data_i.value = dut.data_o.value
            dut.data_valid_i.value = 1
        else:
            dut.data_valid_i.value = 0
        yield RisingEdge(dut.clk_i)
